"""
リクエスト/レスポンスモデル

SOLID原則に基づく設計:
- SRP: 各モデルは単一のリクエスト種別を担当
- ISP: PathRequestMixinで共通インターフェースを提供
- DRY: 共通フィールド定義をAnnotatedで抽出
"""

from typing import Annotated

from pydantic import BaseModel, Field


# 共通フィールド定義（DRY原則）
FilePath = Annotated[str, Field(min_length=1, description="ファイルパス")]
FileContent = Annotated[str, Field(description="ファイル内容")]


class PathRequestMixin(BaseModel):
    """パスを持つリクエストの共通インターフェース（ISP）"""

    path: FilePath


class SaveFileRequest(PathRequestMixin):
    """ファイル保存リクエスト"""

    content: FileContent


class CreateDirectoryRequest(PathRequestMixin):
    """フォルダ作成リクエスト"""

    pass


class MoveItemRequest(BaseModel):
    """ファイル/フォルダ移動リクエスト"""

    source: FilePath
    destination: FilePath



"""
WebSocket接続管理とブロードキャスト機能

SRP: WebSocket通信のみを担当
"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Set

from fastapi import WebSocket

if TYPE_CHECKING:
    pass


class WebSocketManager:
    """WebSocket接続の管理とブロードキャストを担当"""

    def __init__(self) -> None:
        self._clients: Set[WebSocket] = set()

    @property
    def clients(self) -> Set[WebSocket]:
        return self._clients

    @property
    def has_clients(self) -> bool:
        return len(self._clients) > 0

    def add(self, client: WebSocket) -> None:
        self._clients.add(client)

    def remove(self, client: WebSocket) -> None:
        self._clients.discard(client)

    async def broadcast(self, message: dict[str, Any]) -> None:
        """全クライアントにメッセージを送信"""
        if not self._clients:
            return

        message_json = json.dumps(message)
        disconnected: list[WebSocket] = []

        for client in self._clients:
            try:
                await client.send_text(message_json)
            except Exception:
                disconnected.append(client)

        # 切断されたクライアントを削除
        for client in disconnected:
            self.remove(client)


# シングルトンインスタンス
ws_manager = WebSocketManager()



"""
MDV Server Tests
"""

import os
import tempfile
from pathlib import Path

import pytest
from fastapi.testclient import TestClient

from mdv.server import app
from mdv.state import state


@pytest.fixture
def temp_dir():
    """テスト用一時ディレクトリ"""
    with tempfile.TemporaryDirectory() as tmpdir:
        # テスト用ファイルを作成
        readme = Path(tmpdir) / "README.md"
        readme.write_text("# Hello\n\nThis is a test.")

        code_file = Path(tmpdir) / "test.py"
        code_file.write_text("print('hello')")

        subdir = Path(tmpdir) / "subdir"
        subdir.mkdir()
        (subdir / "nested.md").write_text("# Nested")

        yield tmpdir


@pytest.fixture
def client(temp_dir):
    """テストクライアント"""
    state.set_root_path(temp_dir)
    state.last_mtime = 0
    state.current_watching_file = None
    return TestClient(app)


class TestAPIEndpoints:
    """API エンドポイントのテスト"""

    def test_index(self, client):
        """/ エンドポイント"""
        response = client.get("/")
        assert response.status_code == 200
        assert "text/html" in response.headers["content-type"]

    def test_get_tree(self, client, temp_dir):
        """/api/tree エンドポイント"""
        response = client.get("/api/tree")
        assert response.status_code == 200

        tree = response.json()
        assert isinstance(tree, list)

        # ファイルが含まれていることを確認
        names = [item["name"] for item in tree]
        assert "README.md" in names
        assert "test.py" in names
        assert "subdir" in names

    def test_get_info(self, client, temp_dir):
        """/api/info エンドポイント"""
        response = client.get("/api/info")
        assert response.status_code == 200

        info = response.json()
        assert "rootPath" in info
        assert "rootName" in info

    def test_get_file_markdown(self, client, temp_dir):
        """/api/file - Markdownファイル"""
        response = client.get("/api/file", params={"path": "README.md"})
        assert response.status_code == 200

        data = response.json()
        assert data["name"] == "README.md"
        assert data["fileType"] == "markdown"
        assert "content" in data
        assert "raw" in data
        assert "<h1" in data["content"]  # HTMLに変換されている

    def test_get_file_code(self, client, temp_dir):
        """/api/file - コードファイル"""
        response = client.get("/api/file", params={"path": "test.py"})
        assert response.status_code == 200

        data = response.json()
        assert data["name"] == "test.py"
        assert data["fileType"] == "code"

    def test_get_file_nested(self, client, temp_dir):
        """/api/file - ネストされたファイル"""
        response = client.get("/api/file", params={"path": "subdir/nested.md"})
        assert response.status_code == 200

        data = response.json()
        assert data["name"] == "nested.md"

    def test_get_file_not_found(self, client):
        """/api/file - 存在しないファイル"""
        response = client.get("/api/file", params={"path": "nonexistent.md"})
        assert response.status_code == 404

    def test_get_file_path_traversal(self, client, temp_dir):
        """/api/file - パストラバーサル防止"""
        response = client.get("/api/file", params={"path": "../../../etc/passwd"})
        assert response.status_code in [403, 404]

    def test_save_file(self, client, temp_dir):
        """/api/file POST - ファイル保存"""
        new_content = "# Updated\n\nNew content here."
        response = client.post(
            "/api/file",
            json={"path": "README.md", "content": new_content}
        )
        assert response.status_code == 200

        # 実際にファイルが更新されたか確認
        readme = Path(temp_dir) / "README.md"
        assert readme.read_text() == new_content

    def test_delete_file(self, client, temp_dir):
        """/api/file DELETE - ファイル削除"""
        # テスト用ファイルを作成
        test_file = Path(temp_dir) / "to_delete.md"
        test_file.write_text("delete me")

        response = client.delete("/api/file", params={"path": "to_delete.md"})
        assert response.status_code == 200
        assert not test_file.exists()

    def test_create_directory(self, client, temp_dir):
        """/api/mkdir - ディレクトリ作成"""
        response = client.post("/api/mkdir", json={"path": "new_folder"})
        assert response.status_code == 200

        new_dir = Path(temp_dir) / "new_folder"
        assert new_dir.exists()
        assert new_dir.is_dir()

    def test_move_file(self, client, temp_dir):
        """/api/move - ファイル移動"""
        # テスト用ファイルを作成
        source = Path(temp_dir) / "source.md"
        source.write_text("source content")

        response = client.post(
            "/api/move",
            json={"source": "source.md", "destination": "moved.md"}
        )
        assert response.status_code == 200

        assert not source.exists()
        assert (Path(temp_dir) / "moved.md").exists()


class TestMarkdownRendering:
    """Markdownレンダリングのテスト"""

    def test_basic_markdown(self, client, temp_dir):
        """基本的なMarkdown変換"""
        md_file = Path(temp_dir) / "test.md"
        md_file.write_text("# Heading\n\n**bold** and *italic*")

        response = client.get("/api/file", params={"path": "test.md"})
        content = response.json()["content"]

        assert "<h1" in content
        assert "<strong>bold</strong>" in content
        assert "<em>italic</em>" in content

    def test_code_block(self, client, temp_dir):
        """コードブロックの変換"""
        md_file = Path(temp_dir) / "code.md"
        md_file.write_text("```python\nprint('hello')\n```")

        response = client.get("/api/file", params={"path": "code.md"})
        content = response.json()["content"]

        assert "<pre>" in content
        assert "<code" in content
        assert "language-python" in content

    def test_yaml_frontmatter(self, client, temp_dir):
        """YAMLフロントマターの処理"""
        md_file = Path(temp_dir) / "frontmatter.md"
        md_file.write_text("---\ntitle: Test\nauthor: Me\n---\n\n# Content")

        response = client.get("/api/file", params={"path": "frontmatter.md"})
        content = response.json()["content"]

        # YAMLがコードブロックとして表示される
        assert "language-yaml" in content

    def test_mermaid_block(self, client, temp_dir):
        """Mermaidブロックの処理"""
        md_file = Path(temp_dir) / "mermaid.md"
        md_file.write_text("```mermaid\ngraph TD\n    A --> B\n```")

        response = client.get("/api/file", params={"path": "mermaid.md"})
        content = response.json()["content"]

        assert "language-mermaid" in content

    def test_task_list(self, client, temp_dir):
        """タスクリストの変換"""
        md_file = Path(temp_dir) / "tasks.md"
        md_file.write_text("- [x] Done\n- [ ] Todo")

        response = client.get("/api/file", params={"path": "tasks.md"})
        content = response.json()["content"]

        assert 'type="checkbox"' in content


class TestFileWatcher:
    """ファイル監視のテスト"""

    def test_set_watching_file(self, temp_dir):
        """監視ファイルの設定"""
        state.set_root_path(temp_dir)
        readme = Path(temp_dir) / "README.md"

        state.set_watching_file(str(readme))

        assert state.current_watching_file == str(readme)
        assert state.last_mtime > 0

    def test_mtime_change_detection(self, temp_dir):
        """mtime変更の検出"""
        state.set_root_path(temp_dir)
        readme = Path(temp_dir) / "README.md"

        state.set_watching_file(str(readme))
        original_mtime = state.last_mtime

        # ファイルを更新
        import time
        time.sleep(0.1)  # mtimeの精度のため
        readme.write_text("# Updated content")

        new_mtime = os.path.getmtime(str(readme))
        assert new_mtime != original_mtime


class TestSecurity:
    """セキュリティテスト"""

    def test_path_traversal_dotdot(self, client):
        """パストラバーサル: ../"""
        response = client.get("/api/file", params={"path": "../secret.txt"})
        assert response.status_code in [403, 404]

    def test_path_traversal_encoded(self, client):
        """パストラバーサル: エンコードされた../"""
        response = client.get("/api/file", params={"path": "..%2F..%2Fetc%2Fpasswd"})
        assert response.status_code in [403, 404]

    def test_absolute_path(self, client):
        """絶対パスのブロック"""
        response = client.get("/api/file", params={"path": "/etc/passwd"})
        assert response.status_code in [403, 404]


class TestUploadDownload:
    """アップロード/ダウンロードのテスト"""

    def test_upload_single_file(self, client, temp_dir):
        """単一ファイルのアップロード"""
        file_content = b"test content"
        response = client.post(
            "/api/upload",
            data={"path": ""},
            files={"files": ("uploaded.txt", file_content, "text/plain")}
        )
        assert response.status_code == 200
        result = response.json()
        assert result["success"] is True
        assert "uploaded.txt" in result["uploaded"]

        # ファイルが作成されたか確認
        uploaded_file = Path(temp_dir) / "uploaded.txt"
        assert uploaded_file.exists()
        assert uploaded_file.read_bytes() == file_content

    def test_upload_to_subdirectory(self, client, temp_dir):
        """サブディレクトリへのアップロード"""
        file_content = b"nested content"
        response = client.post(
            "/api/upload",
            data={"path": "subdir"},
            files={"files": ("nested_upload.txt", file_content, "text/plain")}
        )
        assert response.status_code == 200

        uploaded_file = Path(temp_dir) / "subdir" / "nested_upload.txt"
        assert uploaded_file.exists()

    def test_upload_multiple_files(self, client, temp_dir):
        """複数ファイルのアップロード"""
        response = client.post(
            "/api/upload",
            data={"path": ""},
            files=[
                ("files", ("file1.txt", b"content1", "text/plain")),
                ("files", ("file2.txt", b"content2", "text/plain"))
            ]
        )
        assert response.status_code == 200
        result = response.json()
        assert len(result["uploaded"]) == 2

    def test_download_file(self, client, temp_dir):
        """ファイルのダウンロード"""
        response = client.get("/api/download", params={"path": "README.md"})
        assert response.status_code == 200
        assert "attachment" in response.headers.get("content-disposition", "")

    def test_download_not_found(self, client):
        """存在しないファイルのダウンロード"""
        response = client.get("/api/download", params={"path": "nonexistent.txt"})
        assert response.status_code == 404


class TestMediaEndpoints:
    """メディアエンドポイントのテスト"""

    def test_get_image(self, client, temp_dir):
        """画像ファイル取得"""
        # テスト用PNG画像を作成（最小限の有効なPNG）
        png_header = bytes([
            0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A,  # PNG signature
            0x00, 0x00, 0x00, 0x0D, 0x49, 0x48, 0x44, 0x52,  # IHDR chunk
            0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01,  # 1x1 pixel
            0x08, 0x02, 0x00, 0x00, 0x00, 0x90, 0x77, 0x53,
            0xDE, 0x00, 0x00, 0x00, 0x0C, 0x49, 0x44, 0x41,
            0x54, 0x08, 0xD7, 0x63, 0xF8, 0xFF, 0xFF, 0x3F,
            0x00, 0x05, 0xFE, 0x02, 0xFE, 0xDC, 0xCC, 0x59,
            0xE7, 0x00, 0x00, 0x00, 0x00, 0x49, 0x45, 0x4E,
            0x44, 0xAE, 0x42, 0x60, 0x82
        ])
        img_file = Path(temp_dir) / "test.png"
        img_file.write_bytes(png_header)

        response = client.get("/api/image", params={"path": "test.png"})
        assert response.status_code == 200
        assert "image/png" in response.headers["content-type"]

    def test_get_image_not_image(self, client, temp_dir):
        """画像でないファイルへのリクエスト"""
        response = client.get("/api/image", params={"path": "README.md"})
        assert response.status_code == 400

    def test_get_pdf(self, client, temp_dir):
        """PDFファイル取得"""
        # 最小限のPDFを作成
        pdf_content = b"%PDF-1.4\n1 0 obj<</Type/Catalog>>endobj\ntrailer<</Root 1 0 R>>"
        pdf_file = Path(temp_dir) / "test.pdf"
        pdf_file.write_bytes(pdf_content)

        response = client.get("/api/pdf", params={"path": "test.pdf"})
        assert response.status_code == 200
        assert "application/pdf" in response.headers["content-type"]

    def test_get_pdf_not_pdf(self, client, temp_dir):
        """PDFでないファイルへのリクエスト"""
        response = client.get("/api/pdf", params={"path": "README.md"})
        assert response.status_code == 400


class TestUtilities:
    """ユーティリティ関数のテスト"""

    def test_escape_html(self):
        """HTMLエスケープ"""
        from mdv.rendering import escape_html

        assert escape_html("<script>") == "&lt;script&gt;"
        assert escape_html("a & b") == "a &amp; b"
        assert escape_html("normal") == "normal"

    def test_sanitize_filename(self):
        """ファイル名サニタイズ"""
        from mdv.path_security import sanitize_filename

        assert sanitize_filename("test.txt") == "test.txt"
        assert sanitize_filename("../../../etc/passwd") == "passwd"
        assert sanitize_filename("path/to/file.txt") == "file.txt"

    def test_render_code(self):
        """コードレンダリング"""
        from mdv.rendering import render_code

        result = render_code("print('hello')", "python")
        assert "<pre>" in result
        assert "language-python" in result
        assert "print" in result

    def test_render_text(self):
        """テキストレンダリング"""
        from mdv.rendering import render_text

        result = render_text("plain text")
        assert "<pre" in result
        assert "plain-text" in result
        assert "plain text" in result


class TestEdgeCases:
    """エッジケースのテスト"""

    def test_save_nonexistent_file(self, client, temp_dir):
        """存在しないファイルの保存"""
        response = client.post(
            "/api/file",
            json={"path": "nonexistent.md", "content": "new content"}
        )
        assert response.status_code == 404

    def test_mkdir_existing_directory(self, client, temp_dir):
        """既存ディレクトリへのmkdir"""
        response = client.post("/api/mkdir", json={"path": "subdir"})
        assert response.status_code == 400

    def test_move_to_existing_destination(self, client, temp_dir):
        """既存の移動先への移動"""
        response = client.post(
            "/api/move",
            json={"source": "README.md", "destination": "test.py"}
        )
        assert response.status_code == 400

    def test_delete_directory(self, client, temp_dir):
        """ディレクトリの削除"""
        # サブディレクトリを作成
        new_dir = Path(temp_dir) / "to_delete_dir"
        new_dir.mkdir()
        (new_dir / "file.txt").write_text("content")

        response = client.delete("/api/file", params={"path": "to_delete_dir"})
        assert response.status_code == 200
        assert not new_dir.exists()

    def test_get_file_binary_error(self, client, temp_dir):
        """バイナリファイル読み込み（.txtとしてアクセス）"""
        # バイナリファイルを.txtとして作成
        binary_file = Path(temp_dir) / "binary.txt"
        binary_file.write_bytes(bytes([0x00, 0x01, 0x02, 0xFF, 0xFE]))

        response = client.get("/api/file", params={"path": "binary.txt"})
        # UTF-8としてデコードできない場合は400エラー
        assert response.status_code == 400

    def test_validate_path_for_write(self, client, temp_dir):
        """書き込み用パス検証"""
        # 新規ファイルパスでも親ディレクトリが存在すればOK
        response = client.post("/api/mkdir", json={"path": "new_subdir/nested"})
        assert response.status_code == 200
        assert (Path(temp_dir) / "new_subdir" / "nested").exists()


class TestAppState:
    """AppState クラスのテスト"""

    def test_add_remove_client(self):
        """クライアント追加・削除"""
        from mdv.websocket_manager import WebSocketManager

        ws_manager = WebSocketManager()
        mock_client = object()

        ws_manager.add(mock_client)
        assert mock_client in ws_manager.clients

        ws_manager.remove(mock_client)
        assert mock_client not in ws_manager.clients

    def test_remove_nonexistent_client(self):
        """存在しないクライアントの削除（エラーなし）"""
        from mdv.websocket_manager import WebSocketManager

        ws_manager = WebSocketManager()
        mock_client = object()

        # 存在しないクライアントを削除してもエラーにならない
        ws_manager.remove(mock_client)
        assert mock_client not in ws_manager.clients

    def test_set_watching_file_nonexistent(self, temp_dir):
        """存在しないファイルの監視設定"""
        state.set_root_path(temp_dir)
        nonexistent = str(Path(temp_dir) / "nonexistent.md")

        state.set_watching_file(nonexistent)

        assert state.current_watching_file == nonexistent
        assert state.last_mtime == 0  # OSErrorで0にリセット



"""
アプリケーション状態管理

SRP: アプリケーション状態のみを担当
"""

from __future__ import annotations

import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from .file_types import SKIP_DIRECTORIES


@dataclass
class AppState:
    """アプリケーション状態を管理"""
    root_path: Path = field(default_factory=Path.cwd)
    current_watching_file: Optional[str] = None
    last_mtime: float = 0
    dir_mtimes: dict[str, float] = field(default_factory=dict)

    def set_root_path(self, path: str | Path) -> None:
        self.root_path = Path(path).resolve()
        self._update_dir_mtimes()

    def set_watching_file(self, path: str) -> None:
        """監視対象ファイルを設定"""
        self.current_watching_file = path
        try:
            self.last_mtime = os.path.getmtime(path)
        except OSError:
            self.last_mtime = 0

    def check_file_changed(self) -> bool:
        """監視中のファイルが変更されたかチェック"""
        if not self.current_watching_file:
            return False

        try:
            mtime = os.path.getmtime(self.current_watching_file)
            if mtime != self.last_mtime:
                self.last_mtime = mtime
                return True
        except OSError:
            pass

        return False

    def check_dir_changes(self) -> bool:
        """ディレクトリのmtime変更をチェック（変更があればTrue）"""
        changed = False

        try:
            # ルートディレクトリをチェック
            current_mtime = os.path.getmtime(self.root_path)
            if self.dir_mtimes.get(str(self.root_path)) != current_mtime:
                changed = True

            # 直下のサブディレクトリをチェック
            for entry in self.root_path.iterdir():
                if entry.is_dir() and entry.name not in SKIP_DIRECTORIES:
                    try:
                        current = os.path.getmtime(entry)
                        path_str = str(entry)
                        if path_str not in self.dir_mtimes or self.dir_mtimes[path_str] != current:
                            changed = True
                            break
                    except OSError:
                        pass
        except OSError:
            pass

        if changed:
            self._update_dir_mtimes()

        return changed

    def _update_dir_mtimes(self) -> None:
        """監視対象ディレクトリのmtimeを更新"""
        self.dir_mtimes = {}

        try:
            self.dir_mtimes[str(self.root_path)] = os.path.getmtime(self.root_path)

            for entry in self.root_path.iterdir():
                if entry.is_dir() and entry.name not in SKIP_DIRECTORIES:
                    try:
                        self.dir_mtimes[str(entry)] = os.path.getmtime(entry)
                    except OSError:
                        pass
        except OSError:
            pass


# シングルトンインスタンス
state = AppState()



"""
ファイル監視機能

SRP: ファイル変更監視のみを担当
"""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Callable, Awaitable

if TYPE_CHECKING:
    from .state import AppState
    from .websocket_manager import WebSocketManager


async def file_watcher(
    state: "AppState",
    ws_manager: "WebSocketManager",
    on_file_change: Callable[[str], Awaitable[None]],
    on_tree_change: Callable[[], Awaitable[None]],
    poll_interval: float = 0.3,
    dir_check_frequency: int = 3,
) -> None:
    """ファイル変更を監視（ポーリング方式）

    Args:
        state: アプリケーション状態
        ws_manager: WebSocketマネージャー
        on_file_change: ファイル変更時のコールバック
        on_tree_change: ディレクトリ変更時のコールバック
        poll_interval: ポーリング間隔（秒）
        dir_check_frequency: ディレクトリチェック頻度（N回に1回）
    """
    dir_check_counter = 0

    while True:
        await asyncio.sleep(poll_interval)

        if not ws_manager.has_clients:
            continue

        # ファイル変更チェック
        if state.check_file_changed() and state.current_watching_file:
            await on_file_change(state.current_watching_file)

        # ディレクトリ変更チェック（N回に1回）
        dir_check_counter += 1
        if dir_check_counter >= dir_check_frequency:
            dir_check_counter = 0
            if state.check_dir_changes():
                await on_tree_change()



"""
Markdown / Code / Text レンダリング機能

SRP: レンダリングのみを担当
"""

from __future__ import annotations

import re
from typing import Optional

from markdown_it import MarkdownIt
from mdit_py_plugins.tasklists import tasklists_plugin

from .file_types import FileTypeInfo


# === HTML Utilities ===

def escape_html(text: str) -> str:
    """HTMLエスケープ（XSS対策）"""
    return (
        text
        .replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
        .replace('"', "&quot;")
        .replace("'", "&#x27;")
    )


# === Markdown Preprocessing ===

# YAMLフロントマターのパターン（ファイル先頭の---で囲まれた部分）
_FRONTMATTER_PATTERN = re.compile(r'^---\s*\n(.*?)\n---\s*(\n|$)', re.DOTALL)

# 見出し後のYAMLメタデータブロック（---で囲まれたkey: value形式）
_YAML_BLOCK_PATTERN = re.compile(
    r'(^|\n)(#{1,6}\s+[^\n]+)\n+---\s*\n((?:[a-zA-Z_][a-zA-Z0-9_]*:\s*[^\n]*\n?)+)---\s*(\n|$)',
    re.MULTILINE
)

# Mermaidコードブロックのパターン
_MERMAID_PATTERN = re.compile(r'```mermaid\s*\n(.*?)\n```', re.DOTALL)


def _convert_frontmatter_to_codeblock(content: str) -> str:
    """YAMLフロントマターをコードブロックに変換"""
    match = _FRONTMATTER_PATTERN.match(content)
    if match:
        frontmatter_content = match.group(1)
        rest_of_content = content[match.end():]
        return f"```yaml\n{frontmatter_content}\n```\n{rest_of_content}"
    return content


def _convert_yaml_blocks_to_codeblocks(content: str) -> str:
    """見出し後のYAMLメタデータブロックをコードブロックに変換"""
    def replace_yaml_block(match: re.Match) -> str:
        prefix = match.group(1)
        heading = match.group(2)
        yaml_content = match.group(3).rstrip('\n')
        suffix = match.group(4)
        return f"{prefix}{heading}\n\n```yaml\n{yaml_content}\n```{suffix}"

    return _YAML_BLOCK_PATTERN.sub(replace_yaml_block, content)


def _protect_mermaid_blocks(content: str) -> tuple[str, list[str]]:
    """Mermaidコードブロックをプレースホルダで保護"""
    mermaid_blocks: list[str] = []

    def replace_mermaid(match: re.Match) -> str:
        mermaid_blocks.append(match.group(1))
        return f"<!--MERMAID_PLACEHOLDER_{len(mermaid_blocks) - 1}-->"

    protected_content = _MERMAID_PATTERN.sub(replace_mermaid, content)
    return protected_content, mermaid_blocks


def _restore_mermaid_blocks(html: str, mermaid_blocks: list[str]) -> str:
    """Mermaidコードブロックを復元"""
    for i, mermaid_code in enumerate(mermaid_blocks):
        placeholder = f"<!--MERMAID_PLACEHOLDER_{i}-->"
        escaped_code = escape_html(mermaid_code)
        mermaid_html = f'<pre><code class="language-mermaid">{escaped_code}</code></pre>'
        html = html.replace(f"<p>{placeholder}</p>", mermaid_html)
        html = html.replace(placeholder, mermaid_html)
    return html


# === Markdown Parser (Singleton) ===

_md_parser: Optional[MarkdownIt] = None


def _get_md_parser() -> MarkdownIt:
    """markdown-it-pyパーサーを取得（遅延初期化）"""
    global _md_parser
    if _md_parser is None:
        _md_parser = MarkdownIt("commonmark", {"html": True, "typographer": True, "breaks": True})
        _md_parser.enable("table")
        _md_parser.enable("strikethrough")
        _md_parser.use(tasklists_plugin)
    return _md_parser


# === Public Rendering Functions ===

def render_markdown(content: str) -> str:
    """マークダウンをHTMLに変換"""
    # 前処理
    content = _convert_frontmatter_to_codeblock(content)
    content = _convert_yaml_blocks_to_codeblocks(content)
    content, mermaid_blocks = _protect_mermaid_blocks(content)

    # パース＆レンダリング
    md = _get_md_parser()
    tokens = md.parse(content)

    # data-line属性を追加
    for token in tokens:
        if token.map and len(token.map) >= 1:
            if token.attrs is None:
                token.attrs = {}
            token.attrs["data-line"] = str(token.map[0])

    html = md.renderer.render(tokens, md.options, {})

    # 後処理
    return _restore_mermaid_blocks(html, mermaid_blocks)


def render_code(content: str, lang: Optional[str] = None) -> str:
    """コードをシンタックスハイライト用HTMLに変換"""
    escaped = escape_html(content)
    lang_class = f"language-{lang}" if lang else ""
    return f'<pre><code class="{lang_class}">{escaped}</code></pre>'


def render_text(content: str) -> str:
    """プレーンテキストをHTMLに変換"""
    escaped = escape_html(content)
    return f'<pre class="plain-text">{escaped}</pre>'


def render_file_content(content: str, file_info: FileTypeInfo) -> str:
    """ファイルタイプに応じてコンテンツをレンダリング"""
    if file_info.type == "markdown":
        return render_markdown(content)
    if file_info.type == "code":
        return render_code(content, file_info.lang)
    return render_text(content)



"""
ファイルタイプの定義

SOLID/DRY原則に基づいた設計:
- OCP: FILE_TYPE_REGISTRY に追加するだけで新しいファイルタイプをサポート
- DRY: 同じタイプの拡張子をグループ化して重複を排除
"""

from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class FileTypeInfo:
    """ファイルタイプ情報"""

    type: str  # 'markdown', 'text', 'code', 'image', etc.
    icon: str
    lang: Optional[str] = None


# ファイルタイプ定義レジストリ
# 新しいタイプを追加する場合はここにエントリを追加するだけでOK (OCP)
FILE_TYPE_REGISTRY: tuple[tuple[FileTypeInfo, tuple[str, ...]], ...] = (
    # Markdown
    (FileTypeInfo("markdown", "markdown"), (".md", ".markdown")),
    # Text
    (FileTypeInfo("text", "text"), (".txt", ".text", ".rst", ".log")),
    # Config - JSON
    (FileTypeInfo("code", "json", "json"), (".json",)),
    # Config - YAML
    (FileTypeInfo("code", "yaml", "yaml"), (".yaml", ".yml")),
    # Config - TOML
    (FileTypeInfo("code", "toml", "toml"), (".toml",)),
    # Config - INI
    (FileTypeInfo("code", "config", "ini"), (".ini", ".cfg", ".conf")),
    # Config - ENV
    (FileTypeInfo("code", "config", "bash"), (".env",)),
    # Python
    (FileTypeInfo("code", "python", "python"), (".py",)),
    # JavaScript
    (FileTypeInfo("code", "javascript", "javascript"), (".js",)),
    # TypeScript
    (FileTypeInfo("code", "typescript", "typescript"), (".ts",)),
    # React
    (FileTypeInfo("code", "react", "javascript"), (".jsx",)),
    (FileTypeInfo("code", "react", "typescript"), (".tsx",)),
    # HTML
    (FileTypeInfo("code", "html", "html"), (".html",)),
    # CSS
    (FileTypeInfo("code", "css", "css"), (".css",)),
    (FileTypeInfo("code", "css", "scss"), (".scss",)),
    (FileTypeInfo("code", "css", "less"), (".less",)),
    # Java
    (FileTypeInfo("code", "java", "java"), (".java",)),
    # C/C++
    (FileTypeInfo("code", "c", "c"), (".c", ".h")),
    (FileTypeInfo("code", "cpp", "cpp"), (".cpp", ".hpp")),
    # Go
    (FileTypeInfo("code", "go", "go"), (".go",)),
    # Rust
    (FileTypeInfo("code", "rust", "rust"), (".rs",)),
    # Ruby
    (FileTypeInfo("code", "ruby", "ruby"), (".rb",)),
    # PHP
    (FileTypeInfo("code", "php", "php"), (".php",)),
    # Swift
    (FileTypeInfo("code", "swift", "swift"), (".swift",)),
    # Kotlin
    (FileTypeInfo("code", "kotlin", "kotlin"), (".kt",)),
    # Shell
    (FileTypeInfo("code", "shell", "bash"), (".sh", ".bash", ".zsh")),
    # SQL
    (FileTypeInfo("code", "database", "sql"), (".sql",)),
    # XML
    (FileTypeInfo("code", "xml", "xml"), (".xml",)),
    # GraphQL
    (FileTypeInfo("code", "graphql", "graphql"), (".graphql",)),
    # Vue
    (FileTypeInfo("code", "vue", "html"), (".vue",)),
    # Svelte
    (FileTypeInfo("code", "svelte", "html"), (".svelte",)),
    # Jinja2 Templates
    (FileTypeInfo("code", "jinja2", "django"), (".j2", ".jinja", ".jinja2")),
    # Images (DRY: 1つのFileTypeInfoで8拡張子をカバー)
    (
        FileTypeInfo("image", "image"),
        (".png", ".jpg", ".jpeg", ".gif", ".svg", ".webp", ".ico", ".bmp"),
    ),
    # PDF
    (FileTypeInfo("pdf", "pdf"), (".pdf",)),
    # Video (DRY: 1つのFileTypeInfoで6拡張子をカバー)
    (FileTypeInfo("video", "video"), (".mp4", ".webm", ".mov", ".avi", ".mkv", ".m4v")),
    # Audio (DRY: 1つのFileTypeInfoで6拡張子をカバー)
    (FileTypeInfo("audio", "audio"), (".mp3", ".wav", ".ogg", ".flac", ".m4a", ".aac")),
    # Archive
    (FileTypeInfo("archive", "archive"), (".zip", ".tar", ".gz", ".tgz", ".bz2", ".xz", ".rar", ".7z")),
    # Office
    (FileTypeInfo("office", "office"), (".doc", ".docx", ".xls", ".xlsx", ".ppt", ".pptx", ".odt", ".ods", ".odp")),
    # Executable / Installer
    (FileTypeInfo("executable", "executable"), (".exe", ".msi", ".dmg", ".pkg", ".deb", ".rpm", ".app")),
    # Other binary
    (FileTypeInfo("binary", "binary"), (".bin", ".dat", ".iso", ".img")),
)


def _build_file_types_dict() -> dict[str, FileTypeInfo]:
    """レジストリから拡張子マッピング辞書を構築"""
    result: dict[str, FileTypeInfo] = {}
    for file_type_info, extensions in FILE_TYPE_REGISTRY:
        for ext in extensions:
            result[ext] = file_type_info
    return result


# ファイル拡張子とタイプのマッピング (後方互換性のため維持)
FILE_TYPES: dict[str, FileTypeInfo] = _build_file_types_dict()

# サポートする拡張子のセット
SUPPORTED_EXTENSIONS: frozenset[str] = frozenset(FILE_TYPES.keys())

# スキップするディレクトリ
SKIP_DIRECTORIES: frozenset[str] = frozenset(
    [
        "node_modules",
        "__pycache__",
        "venv",
        ".venv",
        ".git",
        "dist",
        "build",
        ".next",
        ".nuxt",
        ".cache",
        "coverage",
        ".pytest_cache",
        ".mypy_cache",
        ".Trash",
        ".Trashes",
        "Library",
        "OneDrive",
        "Dropbox",
        "Google Drive",
        "iCloud Drive",
    ]
)

# スキップするファイル（ゴミファイル）
SKIP_FILES: frozenset[str] = frozenset(
    [
        ".DS_Store",
        ".localized",
        "Thumbs.db",
        "desktop.ini",
    ]
)


def get_file_type(extension: str) -> Optional[FileTypeInfo]:
    """拡張子からファイルタイプ情報を取得"""
    return FILE_TYPES.get(extension.lower())



"""
MDV - Markdown Viewer Server
ファイルツリー表示 + マークダウンプレビュー + ホットリロード

リファクタリング: SOLID/DRY原則に基づき責任を分離
- state.py: アプリケーション状態管理
- websocket_manager.py: WebSocket接続管理
- rendering.py: Markdown/コードレンダリング
- file_tree.py: ファイルツリー構築
- path_security.py: パス検証・セキュリティ
- file_response.py: APIレスポンス生成
- media_streaming.py: メディアストリーミング
- watcher.py: ファイル監視
"""

from __future__ import annotations

import asyncio
import json
import mimetypes
import shutil
import socket
import webbrowser
from contextlib import asynccontextmanager
from pathlib import Path
from typing import List, Optional

import uvicorn
from fastapi import FastAPI, WebSocket, WebSocketDisconnect, Query, HTTPException, Request, UploadFile, File, Form
from fastapi.responses import FileResponse as FastAPIFileResponse
from fastapi.staticfiles import StaticFiles

from .file_response import build_file_response, build_file_update_message
from .file_tree import build_file_tree
from .file_types import get_file_type
from .media_streaming import create_streaming_response
from .path_security import validate_path, validate_path_for_write, sanitize_filename
from .models import SaveFileRequest, CreateDirectoryRequest, MoveItemRequest
from .state import state
from .watcher import file_watcher
from .websocket_manager import ws_manager


# === Broadcast Functions ===

async def broadcast_tree_update() -> None:
    """全クライアントにファイルツリー更新を通知"""
    try:
        tree = build_file_tree(state.root_path, state.root_path)
        await ws_manager.broadcast({"type": "tree_update", "tree": tree})
    except Exception as e:
        print(f"Error broadcasting tree update: {e}")


async def broadcast_file_update(file_path: str) -> None:
    """全クライアントにファイル更新を通知"""
    try:
        message = build_file_update_message(file_path, Path(file_path))
        if message:
            await ws_manager.broadcast(message)
    except Exception as e:
        print(f"Error broadcasting update: {e}")


# === FastAPI Application ===

@asynccontextmanager
async def lifespan(app: FastAPI):
    """アプリケーションのライフサイクル管理"""
    asyncio.create_task(
        file_watcher(
            state=state,
            ws_manager=ws_manager,
            on_file_change=broadcast_file_update,
            on_tree_change=broadcast_tree_update,
        )
    )
    print("File watcher started (polling mode)")
    yield
    print("Server shutting down")


app = FastAPI(title="MDV - Markdown Viewer", lifespan=lifespan)


# === Static Files ===

static_dir = Path(__file__).parent / "static"


@app.get("/")
async def index() -> FastAPIFileResponse:
    """メインページ"""
    return FastAPIFileResponse(static_dir / "index.html")


# === API: File Tree ===

@app.get("/api/tree")
async def get_tree() -> list:
    """ファイルツリーを取得（1階層のみ、遅延読み込み対応）"""
    return build_file_tree(state.root_path, state.root_path, max_depth=1)


@app.get("/api/tree/expand")
async def expand_tree(path: str = Query(...)) -> list:
    """指定ディレクトリの子要素を取得（遅延読み込み用）"""
    dir_path = validate_path(path, state.root_path)

    if not dir_path.is_dir():
        raise HTTPException(status_code=400, detail="Not a directory")

    return build_file_tree(dir_path, state.root_path, max_depth=1)


@app.get("/api/info")
async def get_info() -> dict:
    """サーバー情報を取得"""
    return {
        "rootPath": str(state.root_path),
        "rootName": state.root_path.name or str(state.root_path)
    }


@app.post("/api/shutdown")
async def shutdown() -> dict:
    """サーバーをシャットダウン"""
    import os
    import signal

    async def delayed_shutdown():
        await asyncio.sleep(0.5)
        os.kill(os.getpid(), signal.SIGTERM)

    asyncio.create_task(delayed_shutdown())
    return {"success": True, "message": "Server shutting down..."}


# === API: File Operations ===

@app.get("/api/file")
async def get_file(path: str = Query(...)) -> dict:
    """ファイルを取得してレンダリング"""
    file_path = validate_path(path, state.root_path)

    if not file_path.is_file():
        raise HTTPException(status_code=400, detail="Not a file")

    # 監視対象を更新
    state.set_watching_file(str(file_path.resolve()))

    return build_file_response(path, file_path)


@app.post("/api/file")
async def save_file(request: SaveFileRequest) -> dict:
    """ファイルを保存"""
    file_path = validate_path(request.path, state.root_path)

    if not file_path.is_file():
        raise HTTPException(status_code=400, detail="Not a file")

    file_info = get_file_type(file_path.suffix)
    if not file_info or file_info.type == "image":
        raise HTTPException(status_code=400, detail="Cannot edit this file type")

    try:
        file_path.write_text(request.content, encoding="utf-8")
        return {"success": True, "path": request.path}
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Failed to save: {str(e)}")


@app.delete("/api/file")
async def delete_file(path: str = Query(...)) -> dict:
    """ファイルまたはフォルダを削除"""
    file_path = validate_path(path, state.root_path)

    try:
        if file_path.is_dir():
            shutil.rmtree(file_path)
        else:
            file_path.unlink()
        await broadcast_tree_update()
        return {"success": True, "path": path}
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Failed to delete: {str(e)}")


@app.get("/api/download")
async def download_file(path: str = Query(...)) -> FastAPIFileResponse:
    """ファイルをダウンロード"""
    file_path = validate_path(path, state.root_path)

    if not file_path.is_file():
        raise HTTPException(status_code=400, detail="Not a file")

    mime_type, _ = mimetypes.guess_type(str(file_path))
    return FastAPIFileResponse(
        file_path,
        media_type=mime_type or "application/octet-stream",
        filename=file_path.name
    )


# === API: Media Files ===

@app.get("/api/image")
async def get_image(path: str = Query(...)) -> FastAPIFileResponse:
    """画像ファイルを返す"""
    file_path = validate_path(path, state.root_path)

    mime_type, _ = mimetypes.guess_type(str(file_path))
    if not mime_type or not mime_type.startswith("image/"):
        raise HTTPException(status_code=400, detail="Not an image file")

    return FastAPIFileResponse(file_path, media_type=mime_type)


@app.get("/api/pdf")
async def get_pdf(path: str = Query(...)) -> FastAPIFileResponse:
    """PDFファイルを返す"""
    file_path = validate_path(path, state.root_path)

    if file_path.suffix.lower() != ".pdf":
        raise HTTPException(status_code=400, detail="Not a PDF file")

    return FastAPIFileResponse(file_path, media_type="application/pdf")


@app.get("/api/media")
async def get_media(path: str = Query(...), request: Request = None):
    """動画/音声ストリーミング（Range requests対応）"""
    file_path = validate_path(path, state.root_path)
    return create_streaming_response(file_path, request)


# === API: File Management ===

# アップロードファイルサイズ制限（100MB）
MAX_UPLOAD_SIZE = 100 * 1024 * 1024


@app.post("/api/upload")
async def upload_files(
    path: str = Form(""),
    files: List[UploadFile] = File(...)
) -> dict:
    """ファイルをアップロード（複数ファイル対応、100MBまで）"""
    target_dir = validate_path_for_write(path, state.root_path) if path else state.root_path
    target_dir.mkdir(parents=True, exist_ok=True)

    if not target_dir.is_dir():
        raise HTTPException(status_code=400, detail="Target is not a directory")

    uploaded = []
    for file in files:
        if not file.filename:
            continue

        # ファイルサイズチェック
        file.file.seek(0, 2)  # ファイル末尾に移動
        file_size = file.file.tell()
        file.file.seek(0)  # 先頭に戻す
        if file_size > MAX_UPLOAD_SIZE:
            raise HTTPException(
                status_code=413,
                detail=f"File '{file.filename}' exceeds maximum size of 100MB"
            )

        filename = sanitize_filename(file.filename)
        dest_path = target_dir / filename

        try:
            with open(dest_path, "wb") as f:
                shutil.copyfileobj(file.file, f)
            uploaded.append(filename)
        except Exception as e:
            raise HTTPException(status_code=500, detail=f"Failed to upload {filename}: {str(e)}")

    if uploaded:
        await broadcast_tree_update()

    return {"success": True, "uploaded": uploaded}


@app.post("/api/mkdir")
async def create_directory(request: CreateDirectoryRequest) -> dict:
    """新規フォルダを作成"""
    dir_path = validate_path_for_write(request.path, state.root_path)

    if dir_path.exists():
        raise HTTPException(status_code=400, detail="Directory already exists")

    try:
        dir_path.mkdir(parents=True, exist_ok=True)
        await broadcast_tree_update()
        return {"success": True, "path": request.path}
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Failed to create directory: {str(e)}")


@app.post("/api/move")
async def move_item(request: MoveItemRequest) -> dict:
    """ファイル/フォルダを移動またはリネーム"""
    source_path = validate_path(request.source, state.root_path)
    dest_path = validate_path_for_write(request.destination, state.root_path)

    if dest_path.exists():
        raise HTTPException(status_code=400, detail="Destination already exists")

    try:
        shutil.move(str(source_path), str(dest_path))
        await broadcast_tree_update()
        return {"success": True, "source": request.source, "destination": request.destination}
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Failed to move: {str(e)}")


# === WebSocket ===

@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket) -> None:
    """WebSocket接続を管理"""
    await websocket.accept()
    ws_manager.add(websocket)

    try:
        while True:
            data = await websocket.receive_text()
            message = json.loads(data)

            if message.get("type") == "watch":
                try:
                    file_path = validate_path(message.get("path", ""), state.root_path)
                    state.set_watching_file(str(file_path.resolve()))
                except HTTPException:
                    pass  # 無効なパスは無視

    except WebSocketDisconnect:
        ws_manager.remove(websocket)
    except Exception as e:
        print(f"WebSocket error: {e}")
        ws_manager.remove(websocket)


# === Static Files Mount ===

if static_dir.exists():
    app.mount("/static", StaticFiles(directory=static_dir), name="static")


# === Server Startup ===

def find_available_port(start_port: int, max_attempts: int = 100) -> int:
    """利用可能なポートを探す"""
    for offset in range(max_attempts):
        port = start_port + offset
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(("localhost", port))
                return port
        except OSError:
            continue

    raise RuntimeError(f"No available port found in range {start_port}-{start_port + max_attempts}")


def start_server(
    root_path: str = ".",
    port: int = 8642,
    open_browser: bool = True,
    initial_file: Optional[str] = None,
) -> None:
    """サーバーを起動"""
    state.set_root_path(root_path)

    if not state.root_path.exists():
        print(f"Error: Path does not exist: {state.root_path}")
        return

    try:
        actual_port = find_available_port(port)
        if actual_port != port:
            print(f"Port {port} is in use, using {actual_port} instead")
    except RuntimeError as e:
        print(f"Error: {e}")
        return

    print(f"Serving: {state.root_path}")
    print(f"URL: http://localhost:{actual_port}")

    if open_browser:
        import threading
        from urllib.parse import quote

        url = f"http://localhost:{actual_port}"
        if initial_file:
            url += f"?file={quote(initial_file)}"

        def open_browser_delayed():
            import time
            time.sleep(0.5)
            webbrowser.open(url)

        threading.Thread(target=open_browser_delayed, daemon=True).start()

    try:
        config = uvicorn.Config(app, host="127.0.0.1", port=actual_port, log_level="warning")
        server = uvicorn.Server(config)
        asyncio.run(server.serve())
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    start_server()



#!/usr/bin/env python3
"""
MDV - Markdown Viewer CLI
どこからでも呼び出せるマークダウンビューア
"""

from __future__ import annotations

import argparse
import shutil
import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path


# === Process Management (SRP: プロセス管理の責任) ===


@dataclass
class ProcessInfo:
    """プロセス情報"""
    pid: str
    port: str
    command: str


class ProcessManager:
    """MDVプロセスの検索・停止を担当（SRP）"""

    @staticmethod
    def get_mdv_processes() -> list[ProcessInfo]:
        """稼働中のMDVサーバープロセスを取得"""
        result = subprocess.run(
            ["lsof", "-i", "-P", "-n"],
            capture_output=True,
            text=True,
        )

        processes = []
        for line in result.stdout.strip().split("\n"):
            if "python" not in line.lower() or "LISTEN" not in line:
                continue

            parts = line.split()
            if len(parts) < 9:
                continue

            pid = parts[1]
            process_info = ProcessManager._get_process_info(pid, parts)
            if process_info:
                processes.append(process_info)

        return processes

    @staticmethod
    def _get_process_info(pid: str, parts: list[str]) -> ProcessInfo | None:
        """PIDからMDVプロセス情報を取得"""
        cmd_result = subprocess.run(
            ["ps", "-p", pid, "-o", "command="],
            capture_output=True,
            text=True,
        )
        cmd = cmd_result.stdout.strip()

        if "mdv" not in cmd.lower():
            return None

        port = ProcessManager._extract_port(parts)
        display_cmd = cmd[:80] + "..." if len(cmd) > 80 else cmd

        return ProcessInfo(pid=pid, port=port, command=display_cmd)

    @staticmethod
    def _extract_port(parts: list[str]) -> str:
        """lsof出力からポート番号を抽出"""
        port_info = parts[8] if len(parts) > 8 else ""
        if ":" in port_info:
            return port_info.split(":")[-1].split("->")[0]
        return ""

    @staticmethod
    def kill_process(pid: str) -> bool:
        """指定PIDのプロセスを停止（成功時True）"""
        result = subprocess.run(["kill", pid], capture_output=True)
        return result.returncode == 0


# === Server Commands (SRP: サーバー操作コマンドの責任) ===


def list_servers() -> int:
    """稼働中のMDVサーバーを一覧表示"""
    processes = ProcessManager.get_mdv_processes()

    if not processes:
        print("稼働中のMDVサーバーはありません")
        return 0

    print(f"稼働中のMDVサーバー: {len(processes)}件")
    print("-" * 60)
    print(f"{'PID':<8} {'Port':<8} {'Command'}")
    print("-" * 60)

    for proc in processes:
        print(f"{proc.pid:<8} {proc.port:<8} {proc.command}")

    print("-" * 60)
    print("\n停止: mdv -k -a (全停止) / mdv -k <PID> (個別停止)")
    return 0


def kill_servers(target: str | None = None, kill_all: bool = False) -> int:
    """MDVサーバーを停止"""
    if target:
        return _kill_single_server(target)

    if not kill_all:
        print("全サーバーを停止するには -a オプションが必要です")
        print("   mdv -k -a     全サーバーを停止")
        print("   mdv -k <PID>  特定のサーバーを停止")
        return 1

    return _kill_all_servers()


def _kill_single_server(pid: str) -> int:
    """特定のPIDのサーバーを停止"""
    if ProcessManager.kill_process(pid):
        print(f"PID {pid} を停止しました")
        return 0
    print(f"PID {pid} の停止に失敗しました")
    return 1


def _kill_all_servers() -> int:
    """全サーバーを停止"""
    processes = ProcessManager.get_mdv_processes()

    if not processes:
        print("稼働中のMDVサーバーはありません")
        return 0

    print(f"{len(processes)}件のMDVサーバーを停止します...")

    killed = sum(1 for proc in processes if _kill_process_with_log(proc))

    print(f"\n完了: {killed}/{len(processes)} 件を停止しました")
    return 0 if killed == len(processes) else 1


def _kill_process_with_log(proc: ProcessInfo) -> bool:
    """プロセスを停止してログ出力（成功時True）"""
    if ProcessManager.kill_process(proc.pid):
        print(f"  PID {proc.pid} (port {proc.port}) を停止")
        return True
    print(f"  PID {proc.pid} の停止に失敗")
    return False


# === PDF Conversion (SRP: PDF変換の責任) ===


_NPX_INSTALL_MESSAGE = (
    "Error: npx (Node.js) is required for PDF conversion\n"
    "Install Node.js: https://nodejs.org/"
)


def _validate_markdown_file(input_path: Path) -> str | None:
    """マークダウンファイルを検証（エラー時はエラーメッセージを返す）"""
    if not input_path.exists():
        return f"Error: File not found: {input_path}"
    if not input_path.is_file():
        return f"Error: Not a file: {input_path}"
    if input_path.suffix.lower() not in [".md", ".markdown"]:
        return f"Error: Not a markdown file: {input_path}"
    return None


def _run_md_to_pdf(input_path: Path) -> tuple[bool, str]:
    """md-to-pdfを実行（成功, エラーメッセージ）"""
    cmd = ["npx", "md-to-pdf", str(input_path)]
    result = subprocess.run(
        cmd,
        capture_output=True,
        text=True,
        cwd=str(input_path.parent),
    )

    if result.returncode != 0:
        if "not found" in result.stderr:
            return False, _NPX_INSTALL_MESSAGE
        return False, f"Error: {result.stderr}"

    return True, ""


def _move_pdf_if_needed(default_output: Path, output_path: Path | None) -> int:
    """必要に応じてPDFを移動"""
    if output_path and output_path != default_output:
        if not default_output.exists():
            print(f"Warning: Expected PDF not found at {default_output}")
            return 1
        output_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.move(str(default_output), str(output_path))
        print(f"PDF saved: {output_path}")
    else:
        print(f"PDF saved: {default_output}")
    return 0


def convert_to_pdf(input_path: Path, output_path: Path | None = None) -> int:
    """MarkdownファイルをPDFに変換（md-to-pdfを使用）"""
    validation_error = _validate_markdown_file(input_path)
    if validation_error:
        print(validation_error)
        return 1

    success, error_message = _run_md_to_pdf(input_path)
    if not success:
        print(error_message)
        return 1

    default_output = input_path.with_suffix(".pdf")
    return _move_pdf_if_needed(default_output, output_path)


def start_viewer(
    path: str = ".",
    port: int = 8642,
    open_browser: bool = True,
) -> None:
    """MDVサーバーを起動"""
    target_path = Path(path).resolve()

    if not target_path.exists():
        print(f"Error: Path does not exist: {target_path}")
        sys.exit(1)

    # ファイルが指定された場合、親ディレクトリをルートにして、そのファイルを開く
    initial_file: str | None = None
    if target_path.is_file():
        initial_file = target_path.name
        target_path = target_path.parent

    from .server import start_server

    start_server(
        root_path=str(target_path),
        port=port,
        open_browser=open_browser,
        initial_file=initial_file,
    )


def create_parser() -> argparse.ArgumentParser:
    """引数パーサーを作成"""
    parser = argparse.ArgumentParser(
        prog="mdv",
        description="MDV - Markdown Viewer with file tree + live preview",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  mdv                    Start viewer in current directory
  mdv /path/to/dir       Start viewer in specified directory
  mdv README.md          Open specific file
  mdv --pdf README.md    Convert markdown to PDF
  mdv -p 3000            Start on port 3000
  mdv -l                 List running servers
  mdv -k -a              Stop all servers
""",
    )

    # サーバー管理オプション
    parser.add_argument(
        "-l", "--list",
        action="store_true",
        help="List running MDV servers",
    )
    parser.add_argument(
        "-k", "--kill",
        nargs="?",
        const="__no_pid__",
        metavar="PID",
        help="Stop server (-k -a for all, -k <PID> for specific)",
    )
    parser.add_argument(
        "-a", "--all",
        action="store_true",
        help="Use with -k to stop all servers",
    )

    # ビューア起動オプション
    parser.add_argument(
        "path",
        nargs="?",
        default=".",
        help="Directory or file path to view (default: current directory)",
    )
    parser.add_argument(
        "-p", "--port",
        type=int,
        default=8642,
        help="Server port (default: 8642)",
    )
    parser.add_argument(
        "--no-browser",
        action="store_true",
        help="Don't open browser automatically",
    )

    # PDF変換オプション
    parser.add_argument(
        "--pdf",
        action="store_true",
        help="Convert markdown file to PDF",
    )
    parser.add_argument(
        "-o", "--output",
        type=str,
        metavar="FILE",
        help="Output PDF file path (default: same name as input with .pdf extension)",
    )

    return parser


def _dispatch_command(args: argparse.Namespace) -> int | None:
    """コマンドをディスパッチ（終了コードを返す、Noneはビューア起動）"""
    if args.list:
        return list_servers()

    if args.kill is not None:
        target = args.kill if args.kill != "__no_pid__" else None
        return kill_servers(target=target, kill_all=args.all)

    if args.pdf:
        input_path = Path(args.path).resolve()
        output_path = Path(args.output).resolve() if args.output else None
        return convert_to_pdf(input_path, output_path)

    return None


def main() -> None:
    """メインエントリーポイント"""
    parser = create_parser()
    args = parser.parse_args()

    exit_code = _dispatch_command(args)
    if exit_code is not None:
        sys.exit(exit_code)

    # デフォルト: ビューア起動
    start_viewer(args.path, args.port, not args.no_browser)


if __name__ == "__main__":
    main()



"""
ファイルツリー構築機能

SRP: ファイルツリー構築のみを担当
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

from .file_types import (
    FILE_TYPES,
    SUPPORTED_EXTENSIONS,
    SKIP_DIRECTORIES,
    SKIP_FILES,
)


def build_file_tree(
    root: Path,
    base_path: Path,
    max_depth: int = 1,
    current_depth: int = 0
) -> list[dict[str, Any]]:
    """ディレクトリツリーを構築

    Args:
        root: 走査するディレクトリ
        base_path: 相対パス計算の基準ディレクトリ
        max_depth: 最大深さ（1=直下のみ、0=無制限）
        current_depth: 現在の深さ（内部用）

    Returns:
        ファイルツリー構造のリスト
    """
    items: list[dict[str, Any]] = []

    try:
        entries = sorted(
            root.iterdir(),
            key=lambda x: (not x.is_dir(), x.name.lower())
        )
    except (PermissionError, OSError, TimeoutError):
        return items

    for entry in entries:
        if entry.name in SKIP_DIRECTORIES or entry.name in SKIP_FILES:
            continue

        rel_path = str(entry.relative_to(base_path))

        if entry.is_dir():
            items.append(_build_directory_node(entry, rel_path, base_path, max_depth, current_depth))
        elif entry.suffix.lower() in SUPPORTED_EXTENSIONS:
            items.append(_build_file_node(entry, rel_path))

    return items


def _build_directory_node(
    entry: Path,
    rel_path: str,
    base_path: Path,
    max_depth: int,
    current_depth: int
) -> dict[str, Any]:
    """ディレクトリノードを構築"""
    # 深さ制限チェック
    if max_depth > 0 and current_depth >= max_depth:
        return {
            "name": entry.name,
            "path": rel_path,
            "type": "directory",
            "children": [],
            "loaded": False,
        }

    children = build_file_tree(entry, base_path, max_depth, current_depth + 1)
    return {
        "name": entry.name,
        "path": rel_path,
        "type": "directory",
        "children": children,
        "loaded": True,
    }


def _build_file_node(entry: Path, rel_path: str) -> dict[str, Any]:
    """ファイルノードを構築"""
    file_info = FILE_TYPES[entry.suffix.lower()]
    return {
        "name": entry.name,
        "path": rel_path,
        "type": "file",
        "fileType": file_info.type,
        "icon": file_info.icon,
        "lang": file_info.lang,
    }



"""
ファイルレスポンス生成機能

SRP: APIレスポンス生成のみを担当
OCP: 新しいファイルタイプはハンドラを追加するだけで対応可能
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Optional

from fastapi import HTTPException

from .file_types import FileTypeInfo, get_file_type
from .rendering import render_file_content


# === File Response Handlers ===

def _create_media_response(
    path: str,
    file_path: Path,
    file_info: FileTypeInfo,
    url_template: str
) -> dict[str, Any]:
    """メディアファイル（画像・PDF・動画・音声）のレスポンスを生成"""
    return {
        "path": path,
        "name": file_path.name,
        "fileType": file_info.type,
        url_template: f"/api/{file_info.type}?path={path}" if file_info.type != "video" and file_info.type != "audio" else f"/api/media?path={path}",
    }


def _create_text_response(
    path: str,
    file_path: Path,
    file_info: FileTypeInfo,
    content: str
) -> dict[str, Any]:
    """テキスト系ファイルのレスポンスを生成"""
    html_content = render_file_content(content, file_info)
    return {
        "path": path,
        "name": file_path.name,
        "content": html_content,
        "raw": content,
        "fileType": file_info.type,
        "lang": file_info.lang,
    }


# ファイルタイプとURLキーのマッピング
_MEDIA_URL_KEYS = {
    "image": "imageUrl",
    "pdf": "pdfUrl",
    "video": "mediaUrl",
    "audio": "mediaUrl",
}


def build_file_response(path: str, file_path: Path) -> dict[str, Any]:
    """ファイルパスからAPIレスポンスを生成

    Args:
        path: 相対パス
        file_path: 絶対ファイルパス

    Returns:
        APIレスポンス用辞書

    Raises:
        HTTPException: サポートされていないファイルタイプの場合
    """
    file_info = get_file_type(file_path.suffix)
    if not file_info:
        raise HTTPException(status_code=400, detail="Unsupported file type")

    # バイナリ系ファイルの場合（表示のみ）
    if file_info.type in ("archive", "office", "executable", "binary"):
        return {
            "path": path,
            "name": file_path.name,
            "fileType": file_info.type,
        }

    # メディアファイルの場合
    url_key = _MEDIA_URL_KEYS.get(file_info.type)
    if url_key:
        url_path_map = {
            "image": "/api/image",
            "pdf": "/api/pdf",
            "video": "/api/media",
            "audio": "/api/media",
        }
        url_path = url_path_map.get(file_info.type, "/api/media")
        return {
            "path": path,
            "name": file_path.name,
            "fileType": file_info.type,
            url_key: f"{url_path}?path={path}",
        }

    # テキスト系ファイルの場合
    try:
        content = file_path.read_text(encoding="utf-8")
    except UnicodeDecodeError:
        raise HTTPException(status_code=400, detail="Cannot read binary file as text")

    return _create_text_response(path, file_path, file_info, content)


def build_file_update_message(file_path: str, path: Path) -> Optional[dict[str, Any]]:
    """ファイル更新通知用のメッセージを生成

    Args:
        file_path: ファイルパス文字列
        path: Pathオブジェクト

    Returns:
        WebSocket通知用のメッセージ、またはNone
    """
    file_info = get_file_type(path.suffix)
    if not file_info:
        return None

    if file_info.type == "image":
        return {
            "type": "file_update",
            "path": file_path,
            "fileType": "image",
            "reload": True,
        }

    try:
        content = path.read_text(encoding="utf-8")
    except (UnicodeDecodeError, OSError):
        return None

    html_content = render_file_content(content, file_info)
    return {
        "type": "file_update",
        "path": file_path,
        "content": html_content,
        "raw": content,
        "fileType": file_info.type,
    }



"""
メディアストリーミング機能

SRP: 動画/音声のストリーミングのみを担当
"""

from __future__ import annotations

import mimetypes
import re
from pathlib import Path
from typing import Generator, Optional

from fastapi import HTTPException, Request
from fastapi.responses import StreamingResponse


# ストリーミングのチャンクサイズ (1MB)
CHUNK_SIZE = 1024 * 1024


def _parse_range_header(range_header: str) -> tuple[Optional[int], Optional[int]]:
    """Range ヘッダーをパース"""
    match = re.match(r"bytes=(\d*)-(\d*)", range_header)
    if not match:
        return None, None

    start = int(match.group(1)) if match.group(1) else None
    end = int(match.group(2)) if match.group(2) else None
    return start, end


def _stream_range(file_path: Path, start: int, content_length: int) -> Generator[bytes, None, None]:
    """指定範囲のファイルをストリーミング"""
    with open(file_path, "rb") as f:
        f.seek(start)
        remaining = content_length

        while remaining > 0:
            chunk_size = min(CHUNK_SIZE, remaining)
            chunk = f.read(chunk_size)
            if not chunk:
                break
            remaining -= len(chunk)
            yield chunk


def _stream_file(file_path: Path) -> Generator[bytes, None, None]:
    """ファイル全体をストリーミング"""
    with open(file_path, "rb") as f:
        while chunk := f.read(CHUNK_SIZE):
            yield chunk


def create_streaming_response(
    file_path: Path,
    request: Optional[Request] = None
) -> StreamingResponse:
    """メディアファイルのストリーミングレスポンスを生成

    Args:
        file_path: ファイルパス
        request: HTTPリクエスト（Rangeヘッダー取得用）

    Returns:
        StreamingResponse

    Raises:
        HTTPException: ファイルが存在しないか、Range指定が不正な場合
    """
    if not file_path.is_file():
        raise HTTPException(status_code=400, detail="Not a file")

    file_size = file_path.stat().st_size
    mime_type, _ = mimetypes.guess_type(str(file_path))
    mime_type = mime_type or "application/octet-stream"

    range_header = request.headers.get("range") if request else None

    # Range指定がある場合
    if range_header:
        start, end = _parse_range_header(range_header)

        if start is not None:
            if start >= file_size:
                raise HTTPException(status_code=416, detail="Range not satisfiable")

            end = min(end, file_size - 1) if end is not None else file_size - 1
            content_length = end - start + 1

            return StreamingResponse(
                _stream_range(file_path, start, content_length),
                status_code=206,
                media_type=mime_type,
                headers={
                    "Content-Range": f"bytes {start}-{end}/{file_size}",
                    "Accept-Ranges": "bytes",
                    "Content-Length": str(content_length),
                }
            )

    # Range指定なしの場合は全体を返す
    return StreamingResponse(
        _stream_file(file_path),
        media_type=mime_type,
        headers={
            "Accept-Ranges": "bytes",
            "Content-Length": str(file_size),
        }
    )



"""
パス検証・セキュリティ機能

SRP: パスのセキュリティ検証のみを担当
"""

from __future__ import annotations

from pathlib import Path

from fastapi import HTTPException


def validate_path(requested_path: str, root_path: Path) -> Path:
    """パスを検証してセキュアなPathオブジェクトを返す

    Args:
        requested_path: リクエストされたパス
        root_path: 許可されたルートパス

    Returns:
        検証済みのPathオブジェクト

    Raises:
        HTTPException: ファイルが存在しないか、アクセス権限がない場合
    """
    file_path = root_path / requested_path

    if not file_path.exists():
        raise HTTPException(status_code=404, detail="File not found")

    _validate_within_root(file_path, root_path)

    return file_path


def validate_path_for_write(requested_path: str, root_path: Path) -> Path:
    """書き込み用のパス検証（ファイルが存在しなくてもOK）

    Args:
        requested_path: リクエストされたパス
        root_path: 許可されたルートパス

    Returns:
        検証済みのPathオブジェクト

    Raises:
        HTTPException: アクセス権限がない場合
    """
    file_path = root_path / requested_path
    _validate_within_root(file_path, root_path)
    return file_path


def _validate_within_root(file_path: Path, root_path: Path) -> None:
    """パスがルートパス内にあることを確認"""
    try:
        file_path.resolve().relative_to(root_path.resolve())
    except ValueError:
        raise HTTPException(status_code=403, detail="Access denied")


def sanitize_filename(filename: str) -> str:
    """ファイル名をサニタイズ（パス区切り文字を除去）"""
    return Path(filename).name
